import click
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
import pandas as pnd
import rasterio
from shapely.geometry import shape, Point, LineString

from drapery.ops.sample import sample
import surficial
from surficial.ops.shape import densify_linestring
from surficial.cli import defaults, util

@click.command(options_metavar='<options>')
@click.argument('alignment_f', nargs=1, type=click.Path(exists=True), metavar='<alignment_file>')
@click.option('--surface', 'elevation_f', nargs=1, type=click.Path(exists=True), metavar='<surface_file>')
@click.option('--points', 'point_multi_f', type=(click.Path(exists=True), click.STRING), multiple=True, metavar='<point_file> <style>',
              help='Points to project onto profile using a given style')
@click.option('--styles', 'styles_f', nargs=1, type=click.Path(exists=True), metavar='<styles_file>',
              help="JSON file containing plot styles")
@click.option('--label/--no-label', is_flag=True, default=False,
              help="Label features from a given field in the features dataset")
@click.option('--despike/--no-despike', is_flag=True, default=True,
              help="Eliminate elevation up-spikes from the stream profile")
@click.option('--densify', nargs=1, type=click.FLOAT, metavar='<float>',
              help="Densify lines with regularly spaced stations given a value for step in map units")
@click.option('--radius', nargs=1, type=click.FLOAT, default=100, metavar='<float>',
              help="Search radius buffer; points within the buffer will display in profile")
@click.option('--invert/--no-invert', is_flag=True, default=True,
              help="Invert the x-axis")
@click.option('-e', '--exaggeration', nargs=1, type=click.INT, default=100, metavar='<int>',
              help="Vertical exaggeration of the profile")
@click.pass_context
def profile(ctx, alignment_f, elevation_f, point_multi_f, styles_f, label, despike, densify, radius, invert, exaggeration):
    """
    Plots a long profile

    \b
    Example:
    surficial profile stream_ln.shp --surface elevation.tif --points feature_pt.shp features --points terrace_pt.shp terrace --styles styles.json

    """
    _, alignment_crs, lines = util.read_geometries(alignment_f)
    base_crs, crs_status = util.check_crs(alignment_crs)
    if crs_status != 'success':
        raise click.BadParameter('{} is {}'.format(alignment_f, crs_status))
    unit = base_crs.GetAttrValue('unit')

    if densify:
        lines = [densify_linestring(line, step=densify) for line in lines]

    if elevation_f:
        with rasterio.open(elevation_f) as elevation_src:
            lines = [LineString(sample(elevation_src, line.coords)) for line in lines]
        
    alignment = surficial.Alignment(lines)
    edge_addresses = alignment.edge_addresses(alignment.outlet())
    vertices = alignment.vertices()

    if despike:
        #vertices = surficial.remove_spikes(vertices)
        vertices = surficial.remove_spikes_graph(alignment)

    # -----------
    # PLOTTING
    # -----------    
    styles = defaults.styles.copy()
    if styles_f:
        user_styles = util.load_style(styles_f)
        styles.update(user_styles)

    handles = []
    fig = plt.figure()
    ax = fig.add_subplot(111)

    profile_verts = [list(zip(edge['route_m'], edge['z'])) for _, edge in vertices.groupby('edge')]
    profile_lines = LineCollection(profile_verts, **styles.get('alignment'))
    ax.add_collection(profile_lines)

    if despike:
        despiked_verts = [list(zip(edge['route_m'], edge['zmin'])) for _, edge in vertices.groupby('edge')]
        despiked_lines = LineCollection(despiked_verts, **styles.get('despiked'))
        ax.add_collection(despiked_lines)

    for point_f, style_key in point_multi_f:
        point_type, point_crs, point_geoms = util.read_geometries(point_f)

        _, crs_status = util.check_crs(point_crs, base_crs=base_crs)
        if crs_status != 'success':
            if crs_status == 'unprojected':
                raise click.BadParameter('{} is unprojected'.format(point_f))
            else:
                msg = 'CRS of {} differs from the CRS of the alignment {}'.format(point_f, alignment_f)
                click.echo(msg)

        if point_type == 'Point':
            with rasterio.open(elevation_f) as elevation_src:
                point_geoms = [Point(sample(elevation_src, [(point.x, point.y)])) for point in point_geoms]
        elif point_type == '3D Point':
            point_geoms = [shape(point['geometry']) for point in point_geoms]
        hits = surficial.points_to_edge_addresses(alignment, point_geoms, radius=radius, reverse=False)
        addresses = surficial.rebase_addresses(hits, edge_addresses)

        # ---------------------------
        # TESTING A ROLLING STATISTIC        
        if style_key == 'terrace':
            means = surficial.rolling_mean(addresses)
            terrace_pts = [list(zip(edge['route_m'], edge['zmean'])) for _, edge in means.groupby('edge')]
            terrace_lines = LineCollection(terrace_pts, **styles.get('mean'))
            ax.add_collection(terrace_lines)
            handles.append(terrace_lines)

            surficial.difference(vertices, means)
        #----------------------------
        # TEST ROLLING
        #surficial.roll_down(alignment, 1, 2, 10)

        #----------------------------
        # TEST EDGE_ADDRESS_TO_XYZ
        #location = surficial.edge_address_to_point(alignment, (5,0),100)
        #print(location)

        #----------------------------
        # TEST DAM IDENTIFICATION
        #surficial.identify_dams(vertices, graph.edges())

        if 'left' and 'right' in styles.get(style_key):
            pts_left, = ax.plot(addresses['route_m'][(addresses.d >= 0)], addresses['z'][(addresses.d >= 0)], **styles.get(style_key).get('left'))
            pts_right, = ax.plot(addresses['route_m'][(addresses.d < 0)], addresses['z'][(addresses.d < 0)], **styles.get(style_key).get('right'))
            handles.extend([pts_left, pts_right])        
        else:
            points, = ax.plot(addresses['route_m'], addresses['z'], **styles.get(style_key))
            handles.append(points)

    extents = util.df_extents(vertices, xcol='route_m', ycol='z')
    padx = (extents.maxx - extents.minx)*0.05
    pady = (extents.maxy - extents.miny)*0.05
    ax.set(aspect=exaggeration,
           xlim=(extents.minx - padx, extents.maxx + padx),
           ylim=(extents.miny - pady, extents.maxy + pady),
           xlabel='Distance ({})'.format(unit.lower()),
           ylabel='Elevation ({0}), {1}x v.e.'.format(unit.lower(), exaggeration))
    if invert:
        ax.invert_xaxis()
    handles.extend([profile_lines, despiked_lines])
    plt.legend(handles=handles)
    plt.show()


"""
def annotate_features(ax, features):
    for measure, feature in features.iterrows():
        offset = 10
        ha = 'left'
        va = 'bottom'
        color = 'black'

        verts = [(measure, feature['ELEVATION']), (measure, feature['ELEVATION'] + offset)]
        codes = [Path.MOVETO, Path.LINETO]
        path = Path(verts, codes)
        patch = patches.PathPatch(path, color=color)
        ax.add_patch(patch)
        ax.text(measure, feature['ELEVATION']+offset, feature['LABEL'], color=color, rotation=30, rotation_mode='anchor', ha=ha, va=va)
    return ax
"""



def load_style(styles_f):
    """Load a json file containing the keyword arguments to use for plot styling

    Parameters:
        styles_f: path to json file containing matplotlib style keyword arguments

    Returns:
        styles (dict): dictionary of matplotlib keyword arguments

    """
    import json

    with open(styles_f, 'r') as styles_src:
        styles = json.load(styles_src)
    return styles

def check_crs(source_crs, base_crs=None):
    """Check the crs for correctness

    Parameters:
        source_crs (str): coordinate reference system in well-known text (WKT) format

    Other Parameters:
        base_crs (str): coordinate reference system for comparison

    Returns:
        crs (SpatialReference): coordinate reference system of the source data
        status (str): status message

    """
    from gdal import osr

    crs=osr.SpatialReference(wkt=source_crs)
    if crs.IsProjected and base_crs == None:
        status = 'success'
    elif crs.IsProjected and crs == base_crs:
        status = 'success'
    elif crs.IsProjected and crs != base_crs:
        status = 'unequal'
    else:
        status = 'unprojected'

    return crs, status

def read_geometries(feature_f):
    """Read feature source geometries

    Parameters:
        feature_f: path to the feature data to read

    Returns:
        schema_geometry (str): feature type
        feature_crs (str): feature source crs in well-known text (WKT) format 
        geometries: list of shapely geometries

    """
    import click
    import fiona
    import rasterio
    from shapely.geometry import shape

    with fiona.open(feature_f) as feature_src:
        supported = ['Point', 'LineString', '3D Point', '3D LineString']
        schema_geometry = feature_src.schema['geometry'] 
        try:
            if schema_geometry not in supported:
                raise click.BadParameter('Geometry must be one of: {}'.format(supported))
        except:
            raise click.BadParameter('Unable to obtain schema from {}'.format(feature_f))
        geometries = [shape(feature['geometry']) for feature in feature_src]
        feature_crs = feature_src.crs_wkt
    return schema_geometry, feature_crs, geometries

def df_extents(df, xcol='x', ycol='y'):
    from collections import namedtuple

    Extents = namedtuple('Extents', ['minx', 'miny', 'maxx', 'maxy'])

    extents = Extents(df[xcol].min(), df[ycol].min(), df[xcol].max(), df[ycol].max())
    return extents


import sys

import click
import fiona
from shapely.geometry import shape, mapping

import surficial

@click.command(options_metavar='<options>')
@click.argument('alignment_f', nargs=1, type=click.Path(exists=True), metavar='<alignment_file>')
@click.argument('output_f', nargs=1, type=click.Path(), metavar='<output_file>')
@click.argument('radius', nargs=1, type=click.FLOAT, metavar='<float>')
@click.option('-s', '--source', nargs=1, type=click.INT, metavar='<int>', help="Source node ID")
@click.option('-o', '--outlet', nargs=1, type=click.INT, metavar='<int>', help="Outlet node ID")
@click.pass_context
def buffer(ctx, alignment_f, output_f, radius, source, outlet):
    """
    Buffers a network graph or path within a network graph

    \b
    Example:
    surficial buffer stream_ln.shp buf.shp 100.0 -s 5

    """
    with fiona.open(alignment_f) as alignment_src:
        lines = [shape(line['geometry']) for line in alignment_src]
        source_driver = alignment_src.driver
        source_crs = alignment_src.crs

    # make the graph
    alignment = surficial.Alignment(lines)

    if not outlet:
        outlet = alignment.outlet()
    if not source:
        path = alignment.edges()
    else:
        path = list(alignment.path_edges(source, outlet))
    buf = alignment.edge_buffer(radius, edges=path)

    sink_schema = {
        'geometry': 'Polygon',
        'properties': {'id': 'int'},
    }

    with fiona.open(
        output_f,
        'w',
        driver=source_driver,
        crs=source_crs,
        schema=sink_schema) as sink:
        sink.write({
           	'geometry': mapping(buf),
           	'properties': {'id': 0},
        })

    click.echo('Output written to: {}'.format(output_f))



styles = {
    "alignment": {
        "color": "black",
        "linestyle": "-",
        "linewidth": 1.0,
        "alpha": 0.2,
        "label": "raw alignment"
    },
    "despiked": {
        "color": "blue",
        "linestyle": "-",
        "linewidth": 1.4,
        "alpha": 1.0,
        "label": "de-spiked alignment"
    },
    "line1": {
        "color": "black",
        "linestyle": "-",
        "linewidth": 1.2,
        "alpha": 0.3,
        "label": "line1"
    },
    "line2": {
        "color": "blue",
        "linestyle": "-",
        "linewidth": 1.4,
        "alpha": 1.0,
        "label": "line2"
    },
    "line3": {
        "color": "green",
        "linestyle": "-",
        "linewidth": 1.4,
        "alpha": 1.0,
        "label": "line3"
    },
    "point1": {
        "color": "red",
        "marker": "o",
        "markersize": 6,
        "markeredgecolor": "black",
        "alpha": 1.0,
        "linestyle": "None",
        "label": "point1"
    },
    "point2": {
        "color": "orange",
        "marker": "o",
        "markersize": 4,
        "markeredgecolor": "None",
        "alpha": 0.5,
        "linestyle": "None",
        "label": "point2"
    },
    "point3": {
        "color": "None",
        "marker": "o",
        "markersize": 4,
        "markeredgecolor": "brown",
        "alpha": 1.0,
        "linestyle": "None",
        "label": "point3"
    }
}


import click
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
import pandas as pnd

import surficial
from surficial.cli import defaults, util

@click.command(options_metavar='<options>')
@click.argument('alignment_f', nargs=1, type=click.Path(exists=True), metavar='<alignment_file>')
@click.option('--points', 'point_multi_f', type=(click.Path(exists=True), click.STRING), multiple=True, metavar='<point_file> <style>',
              help='Plot points on the planview map using a given style')
@click.option('--styles', 'styles_f', nargs=1, type=click.Path(exists=True), metavar='<styles_file>',
              help="JSON file containing plot styles")
@click.option('--show-nodes/--hide-nodes', is_flag=True, default=False,
              help="Label network nodes in the alignment")
@click.pass_context
def plan(ctx, alignment_f, point_multi_f, styles_f, show_nodes):
    """
    Plots a planview map

    \b
    Example:
    surficial plan stream_ln.shp --points terrace_pt.shp terrace --points feature_pt.shp features

    """
    _, alignment_crs, lines = util.read_geometries(alignment_f)
    base_crs, crs_status = util.check_crs(alignment_crs)
    if crs_status != 'success':
        raise click.BadParameter('{} is {}'.format(alignment_f, crs_status))
    unit = base_crs.GetAttrValue('unit')

    alignment = surficial.Alignment(lines)
    vertices = alignment.vertices()

    styles = defaults.styles.copy()
    if styles_f:
        user_styles = util.load_style(styles_f)
        styles.update(user_styles)

    handles = []
    fig = plt.figure()
    ax = fig.add_subplot(111)

    edge_lines = [list(zip(edge_data['x'], edge_data['y'])) for _, edge_data in vertices.groupby(pnd.Grouper(key='edge'))]
    edge_collection = LineCollection(edge_lines, **styles.get('despiked'))
    ax.add_collection(edge_collection)

    for point_f, style_key in point_multi_f:
        _, point_crs, point_geoms = util.read_geometries(point_f)

        _, crs_status = util.check_crs(point_crs, base_crs=base_crs)
        if crs_status != 'success':
            if crs_status == 'unprojected':
                raise click.BadParameter('{} is unprojected'.format(point_f))
            else:
                msg = 'CRS of {} differs from the CRS of the alignment {}'.format(point_f, alignment_f)
                click.echo(msg)

        if 'left' and 'right' in styles.get(style_key):
            click.echo("Left and right styling not implemented in plan view; using left style only.")
            points, = ax.plot([p.coords.xy[0] for p in point_geoms], [p.coords.xy[1] for p in point_geoms], **styles.get(style_key).get('left'))
        else:
            points, = ax.plot([p.coords.xy[0] for p in point_geoms], [p.coords.xy[1] for p in point_geoms], **styles.get(style_key))
        handles.append(points)

    if show_nodes:
        nodes = alignment.nodes(data=True)
        node_labels = [node[0] for node in nodes]
        node_points = [node[1]['geom'] for node in nodes]
        node_x = [p.coords[0][0] for p in node_points]
        node_y = [p.coords[0][1] for p in node_points]
        for label, x, y in zip(node_labels, node_x, node_y):
            plt.annotate(
                label,
                xy = (x, y), xytext = (-20, 20),
                textcoords = 'offset points', ha = 'right', va = 'bottom',
                bbox = dict(boxstyle = 'round,pad=0.3', fc = 'yellow', alpha = 0.5),
                arrowprops = dict(arrowstyle = '->', connectionstyle = 'arc3,rad=0'))
        nodes, = ax.plot(node_x, node_y, **styles.get('point1'))
        handles.append(nodes)

    handles.append(edge_collection)

    extents = util.df_extents(vertices, xcol='x', ycol='y')
    padx = (extents.maxx - extents.minx)*0.05
    pady = (extents.maxy - extents.miny)*0.05
    ax.set(aspect=1,
           xlim=(extents.minx - padx, extents.maxx + padx),
           ylim=(extents.miny - pady, extents.maxy + pady),
           xlabel='Easting ({})'.format(unit.lower()),
           ylabel='Northing ({})'.format(unit.lower()))
    plt.legend(handles=handles)
    plt.show()



import warnings
from operator import itemgetter

import networkx as nx
from networkx import DiGraph
import pandas as pnd
from shapely.geometry import Point, MultiLineString

from surficial.ops.shape import measure, linestring_to_vertices, linestring_to_stations, densify_linestring

class Alignment(DiGraph):
    """A directed network graph of LineStrings.

    Alignment is a subclass of networkx.DiGraph and adds methods for addressing points within
    the network. It represents the set of geometries onto which points of interest are
    projected.

    """
    def __init__(self, lines):
        """Construct a directed graph from a set of LineStrings.

        Parameters:
            lines (list of LineString): geometries in the network

        Returns:
            graph (DiGraph): directed network graph

        """
        super().__init__()

        # add the nodes
        endpoints = []
        for line in lines:
            endpoints.append(line.coords[0])
            endpoints.append(line.coords[-1])
        for i, p in enumerate(set(endpoints)):
            self.add_node(i, geom=Point(p))

        # add the edges
        for line in lines:
            from_node = None
            to_node = None
            for n, data in self.nodes(data=True):
                p = data['geom']
                if p.equals(Point(line.coords[0])):
                    from_node = n
                elif p.equals(Point(line.coords[-1])):
                    to_node = n
            self.add_edge(from_node, to_node, geom=line, len=line.length, meas=measure(line))

        if nx.isolates(self):
            warnings.warn("Found isolated nodes, check input geometries using the repair subcommand. Exiting now.")
        if len(list(nx.connected_component_subgraphs(self.to_undirected()))) > 1:
            warnings.warn("Found multiple subgraphs, check input geometries using the repair subcommand. Exiting now.")

    def outlet(self):
        """Return the root node in a directed graph.

        In a stream network this represents the drainage outlet.

        Returns:
            n (int): outlet node ID

        """
        for node in self.nodes():
            if self.out_degree(node) == 0:
                return node

    def edge_addresses(self, outlet, weight='len'):
        """Calculate cost path distances from a given node to each graph edge end node. 

        The DataFrame columns are:

            :edge (tuple): tuple of node identifiers identifying an edge
            :to_node_address (float): the cost path distance between outlet node and edge end node

        Parameters:
            outlet (int): network outlet node ID

        Other Parameters:
            weight (string): name of property to use for weight calculation

        Returns:
            result (DataFrame): edge address information relative to outlet

        """
        addresses = []
        for from_node, to_node, _ in self.edges(data=True):
            to_node_path = self.path_edges(to_node, outlet)
            to_node_dist = self.path_weight(to_node_path, weight)

            from_node_path = self.path_edges(from_node, outlet)
            from_node_dist = self.path_weight(from_node_path, weight)

            addresses.append([(from_node, to_node), from_node_dist, to_node_dist])
        result = pnd.DataFrame(addresses, columns=['edge', 'from_node_address', 'to_node_address'])
        return result

    def edge_buffer(self, radius, edges=None):
        """Return a buffer Polygon around a set of graph edges.

        \b
        Example:
        # get edges along a path, buffer them and make a PolygonPatch for plotting in MPL
        path = list(alignment.path_edges(1, outlet))
        buf = PolygonPatch(alignment.edge_buffer(100.0, edges=path), fc=BLUE, ec=BLUE, alpha=0.5, zorder=2)

        Parameters:
            radius (float): buffer radius

        Other Parameters:
            edges (list of tuples): optional list of edges to buffer

        Returns:
            polygon (MultiLineString): polygon representing the buffered geometries

        """
        if edges is None:
            edges = self.edges()
        geoms = [self[from_node][to_node]['geom'] for (from_node, to_node) in edges]
        polygon = MultiLineString(geoms).buffer(radius)
        return polygon

    def path_edges(self, start, goal, weight=None):
        """Return the set of graph edges making up a shortest path.

        Parameters:
            start (int): starting node ID
            goal (int): goal node ID

        Other Parameters:
            weight (string): name of property to use for weight calculation

        Returns:
            edges (list of tuples): list of edges making up the path

        """
        path = nx.shortest_path(self, start, goal, weight=weight)
        edges = zip(path[:-1], path[1:])

        return edges

    def path_weight(self, edges, weight):
        """Return the path weight of a set of graph edges.

        Parameters:
            edges (list of tuples): list of edges making up the path
            weight (string): name of property to use for weight calculation

        Returns:
            total (float): path weight

        """
        total = 0
        for (from_node, to_node) in edges:
            total += self[from_node][to_node][weight]
        return total

    def station(self, step):
        """Get a dataframe of regularly spaced stations along graph edges.

        The DataFrame columns are:

            :m (float): path distance from the to_node endpoint
            :x (float): x coordinate
            :y (float): y coordinate
            :z (float): z coordinate
            :edge (tuple): pair of graph nodes (from, to)

        Parameters:
            step (float): distance spacing between stations

        Returns:
            stations (DataFrame): point information

        """
        edge_addresses = self.edge_addresses(self.outlet())

        stations = pnd.DataFrame()
        for from_node, to_node, data in self.edges(data=True):
            path = self.path_edges(from_node, self.outlet())
            path_len = self.path_weight(path, 'len')
            line = data['geom']

            end_address = edge_addresses[edge_addresses['edge'] == (from_node, to_node)]
            start = (end_address.iloc[0]['to_node_address'] + line.length) % step

            line_stations = pnd.DataFrame(linestring_to_stations(line, position=start, step=step), columns=['m', 'x', 'y', 'z'])
            line_stations['route_m'] = path_len - line_stations['m']
            line_stations['edge'] = [(from_node, to_node) for station in range(line_stations.shape[0])] 

            if stations.empty:
                stations = line_stations
            else:
                stations = pnd.concat([stations, line_stations], ignore_index=True)

        return stations

    def vertices(self):
        """Get a dataframe of the vertices

        The DataFrame columns are:

            :m (float): distance from the edge endpoint
            :x (float): x coordinate
            :y (float): y coordinate
            :z (float): z coordinate
            :edge (tuple): pair of graph nodes (from, to)

        Returns:
            vertices (DataFrame): point information

        """
        vertices = pnd.DataFrame()
        for from_node, to_node, data in self.edges(data=True):
            path = self.path_edges(from_node, self.outlet())
            path_len = self.path_weight(path, 'len')

            line_vertices = pnd.DataFrame(linestring_to_vertices(data['geom']), columns=['m','x','y','z'])
            line_vertices['route_m'] = path_len - line_vertices['m']
            line_vertices['edge'] = [(from_node, to_node) for vertex in range(line_vertices.shape[0])] 

            if vertices.empty:
                vertices = line_vertices
            else:
                vertices = pnd.concat([vertices, line_vertices], ignore_index=True)

        return vertices

    def intermediate_nodes(self):
        """Return the set of nodes intermediate between leaf and root nodes.

        Returns:
            node_list (list of int): list of all intermediate node ID values

        """
        node_list = [node for node in self.nodes() if self.out_degree(node) > 0 and self.in_degree(node) > 0]
        return node_list



from operator import itemgetter

import networkx as nx
from shapely.geometry import Point, MultiLineString
import pandas as pnd

from surficial.ops.shape import measure, filter_contains, project2d

def points_to_edge_addresses(graph, points, radius=100, edges=None, reverse=False):
    """Locate points by address along the nearest graph edge.

    Returns a DataFrame describing the addresses (projections) of points, within some distance, onto a set of graph edges.
    The DataFrame columns are:

        :m (float): distance along the edge geometry
        :x (float): projected point x coordinate
        :y (float): projected point y coordinate
        :z (float): projected point z coordinate
        :d (float): offset distance, or distance from the point to its projection
        :edge (tuple): tuple of node identifiers identifying an edge 

    Parameters:
        graph (DiGraph): directed network graph
        points (list of Points): points to project

    Other Parameters:
        radius (float): buffer radius
        edges (list of tuples): edge tuples onto which points will be projected, if None then all edges in graph are used
        reverse (bool): reverse vertex ordering

    Returns:
        rows_df (DataFrame): point address information relative to individual edges

    """
    if edges is None:
        edges = graph.edges()
        
    rows = []
    for edge in edges:
        edge_rows = []
        buffer = graph.edge_buffer(radius, edges=[edge])
        pts = filter_contains(points, buffer)
        geom = graph[edge[0]][edge[1]]['geom']
        meas = graph[edge[0]][edge[1]]['meas']
        for p in pts:
            pp = project2d(p, geom, measure=meas)
            # i think i mean to use either d or u below as offset is left or right of the line
            if reverse is True:
                m = geom.length - pp['m']
            else: m = pp['m']
            if m > 0 and m < geom.length:
                edge_rows.append([m, pp['pt'].x, pp['pt'].y, pp['pt'].z, pp['d'], edge])
        rows.extend(sorted(edge_rows, key=itemgetter(0), reverse=False))
    rows_df = pnd.DataFrame(rows, columns=['m', 'x', 'y', 'z', 'd', 'edge'])
    
    return rows_df

def rebase_addresses(point_addresses, edge_addresses):
    """Calculate point distances from a node.

    The DataFrame columns are:

        :route_m (float): distance along the route from the projected point the outlet node
        :m (float): distance along the edge geometry
        :x (float): projected point x coordinate
        :y (float): projected point y coordinate
        :z (float): projected point z coordinate
        :d (float): offset distance, or distance from the point to its projection
        :edge (tuple): tuple of node identifiers identifying an edge
        :to_node_address (float): cost path distance from the edge end node to the outlet node

    Parameters:
        point_addresses (DataFrame): point address information
        edge_addresses (DataFrame): edge address information

    Returns:
        result (DataFrame): point address information relative to an outlet node in a network

    """
    addresses = pnd.merge(point_addresses, edge_addresses, on='edge')
    addresses['route_m'] = addresses['from_node_address'] - addresses['m']  # + addresses['to_node_address']

    return addresses

def remove_spikes(vertices):
    """Remove spikes by calculating an expanding minimum from upstream to downstream

    Adds a DataFrame column, zmin, to hold the despiked z-values.

    Parameters:
        vertices (DataFrame): vertex coordinates

    Returns:
        result (DataFrame): vertex coordinates

    """
    grouped = vertices.groupby('edge')
    zmin = grouped['z'].apply(lambda x: x.expanding().min())
    zmin.name = 'zmin'

    result = pnd.concat([vertices, zmin], axis=1)

    return result

def rolling_mean(points):
    """Calculate a rolling mean on a series of point z values

    Parameters:
        points (DataFrame)

    Returns:
        result (DataFrame)
    """
    grouped = points.groupby('edge')
    means = grouped['z'].apply(lambda x: x.rolling(window=9, win_type='triang', center=True).mean())
    means.name = 'zmean'

    result = pnd.concat([points, means], axis=1)
 
    return result
 
def difference(series1, series2):
    """Calculate the difference between zmin and zmean

    Parameters:
        series1 (DataFrame)
        series2 (DataFrame)

    """
    combined = pnd.concat([series1, series2], axis=0, ignore_index=True)
    grouped = combined.groupby('edge')
    for edge, group in grouped:
        aligned = group.sort_values(by='m')
        aligned_m = aligned.set_index('m')
        filled_series = aligned_m.interpolate(method='values')
        filled_series['diff'] = filled_series['zmean'] - filled_series['zmin']

def roll_down(graph, start, goal, window):
    """Perform an operation on a list of path edges

    Parameters:
        graph (Alignment)
        start (int)
        goal (int)
        window (int)

    """
    vertices = graph.vertices()
    edges = list(graph.path_edges(start, goal))
    for i, edge in enumerate(edges):
        pre_window = pnd.DataFrame()
        post_window = pnd.DataFrame()

        verts = vertices[vertices['edge']==edge]
        if i > 0:
            #pre_window = vertices[vertices['edge']==edges[i-1]].tail(window)
            pre_window = get_pre_window(graph.in_edges(edge[0]), vertices, window, 'z')
        if i <= len(edges)-2:
            post_window = vertices[vertices['edge']==edges[i+1]].head(window)

        if pre_window.empty != True and post_window.empty != True:
            extended = pnd.concat([pre_window, verts, post_window])
        elif pre_window.empty != True:
            extended = pnd.concat([pre_window, verts])
        else:
            extended = pnd.concat([verts, post_window])
        roll = extended.sort_values(by='route_m')

        roll['roll'] = roll['z'].rolling(window=window, win_type='triang', center=True).mean()

        result = roll[roll['edge']==edge]
        print(result)

def get_pre_window(edges, vertices, window, column, statistic='min'):
    """Determine a 'winning' edge where a node has multiple edges

    Parameters:
        edges (list of tuples)
        vertices (DataFrame)
        window (int)
        column (string)
        statistic (string)

    """
    in_window = pnd.DataFrame()
    val = None
    for edge in edges:
        tmp = vertices[vertices['edge']==edge].tail(window)
        if val:
            if tmp[column].min() < val:
                in_window = tmp
        else:
            in_window = tmp
            val = tmp[column].min()
    return in_window


def get_min_neighbor_edge(graph, edge, column='z', direction='up', window=None):
    """Return the neighboring edge having the lowest minimum value

    Parameters:
        graph (Alignment)
        edge (tuple)

    Other Parameters:
        column (string)
        direction (string)
        window (int)

    Returns:
        result (tuple): edge having the lowest minimum value

    """
    vertices = graph.vertices()
    min_edge = None
    min_val = None

    if direction=='up':
        edges = [(i, edge[0]) for i in graph.predecessors(edge[0])]
    else:
        edges = [(edge[1], i) for i in graph.successors(edge[1])]

    if len(edges) > 0:
        for test_edge in edges:
            if window:
                test_df = vertices[vertices['edge']==test_edge].tail(window)
            else:
                test_df = vertices[vertices['edge']==test_edge]
            test_val = test_df[column].min()
            if min_val:
                if test_val < min_val:
                    min_edge = test_edge
                    min_val = test_val
            else:
                min_edge = test_edge
                min_val = test_val

    return min_edge

def extend_edge(graph, edge, window=10, statistic="min"):
    """Extend an edge using vertices from neighboring edges

    Parameters:
        graph (Alignment)
        edge (tuple): edge to extend

    Other Parameters:
        window (int): number of vertices to extend the edge
        statistic (string): function used to determine which edge to use among several

    Returns:
        result (DataFrame): vertices of the edge along with vertices from preceeding and successor edges

    """
    vertices = graph.vertices()
    edge_vertices = vertices[vertices['edge']==edge]

    if statistic == 'min':
        pre_edge = get_min_neighbor_edge(graph, edge, column='z', direction='up', window=window)
        post_edge = get_min_neighbor_edge(graph, edge, column='z', direction='down', window=window)

    pre_window = pnd.DataFrame()
    post_window = pnd.DataFrame()
    if pre_edge:
        pre_window = vertices[vertices['edge']==pre_edge].tail(window)
    if post_edge:
        post_window = vertices[vertices['edge']==post_edge].head(window)
    result = pnd.concat([pre_window, edge_vertices, post_window])

    return result

def remove_spikes_graph(graph, start=None, goal=None, column='z'):
    """Remove spikes from a graph or a subset of edges using an expanding minimum

    Parameters:
        graph (Alignment):

    Other Parameters:
        start (int):
        goal (int):
        column (string):

    """
    if start and goal:
        edges = graph.path_edges(start, goal)
    elif start and not goal:
        edges = graph.path_edges(start, graph.outlet())
    elif goal and not start:
        subgraph = graph.subgraph(nx.ancestors(graph, goal))
        edges = subgraph.edges()
    else:
        edges = graph.edges()

    result = pnd.DataFrame()
    for edge in edges:
        edge_data = extend_edge(graph, edge, window=40)
        edge_data['zmin'] = edge_data[column].expanding().min()
        clip = edge_data[edge_data['edge']==edge]
        result = result.append(clip)

    return result

def edge_address_to_point(graph, edge, m):
    """Return a Point location given an edge address within an Alignment

    Parameters:
        graph (Alignment)
        edge (tuple): tuple identifying the edge
        m (float): distance measure along the edge geometry

    Returns:
        point (Point): address location

    """
    line = graph[edge[0]][edge[1]]['geom']
    point = line.interpolate(m)
    
    return point

def identify_dams(graph, min_grade=1.0, min_drop=1.0, column='zmin'):
    """Identify candidate dam locations

    Iterates over a series of vertices and initiates a cumulative accounting of drop in elevation
    for series of line segments having grade greater than a minimum value. Runs of segments
    having a drop greater than the minimum drop value are identified as candidate dams.

    Parameters:
        graph (Alignment)

    Other Parameters:
        column (string)
        min_grade (float)
        min_drop (float)

    """

    # need to move vertices() into init so i can stop doing alignment.vertices()
    # may want to store calculations in separate dataframes from vertices
    vertices = graph.vertices()
    vertices = surficial.remove_spikes_graph(graph)

    indices = []
    for edge in graph.edges():
        edge_data = extend_edge(graph, edge, window=10)
        idx = None
        drop = 0.0
        #edge_data['rise'] = edge_data['zmin'] - edge_data['zmin'].shift(-1)
        #edge_data['slope'] = edge_data['rise'] / (edge_data['m'] - edge_data['m'].shift(-1))
        #print(edge_data['rise'])
        #print(edge_data['slope'])
        print(edge_data)
        """
        _, last_vertex = next(edge_data.itertuples())  # take first item from row_iterator
        for i, vertex in edge_iterator:
            rise = vertex['z'] - last_vertex['z']
            run = vertex['m'] - last_vertex['m']
            slope = rise / run
            print(slope)
        """



import sys

import click
import fiona
from shapely.geometry import Point, LineString, shape, mapping

import surficial


@click.command(options_metavar='<options>')
@click.argument('alignment_f', nargs=1, type=click.Path(exists=True), metavar='<alignment_file>')
@click.argument('output_f', nargs=1, type=click.Path(), metavar='<output_file>')
@click.argument('step', nargs=1, type=click.FLOAT, metavar='<float>')
@click.pass_context
def station(ctx, alignment_f, output_f, step):
    """
    Creates a series of evenly spaced stations

    \b
    Example:
    surficial station stream_ln.shp station_pt.shp 20

    """
    with fiona.open(alignment_f) as alignment_src:
        lines = [shape(line['geometry']) for line in alignment_src]
        source_driver = alignment_src.driver
        source_crs = alignment_src.crs
        source_schema = alignment_src.schema

        alignment = surficial.Alignment(lines)
        vertices = alignment.station(step)

    sink_schema = {
        'geometry': 'Point',
        'properties': {'id': 'int', 'station': 'float', 'from_node': 'int', 'to_node': 'int'},
    }

    with fiona.open(
        output_f,
        'w',
        driver=source_driver,
        crs=source_crs,
        schema=sink_schema) as sink:
            for i, row in vertices.iterrows():
                if row['z'] is not None:
                    geom = Point(row['x'], row['y'], row['z'])
                else:
                    geom = Point(row['x'], row['y'])
                click.echo("Writing id: {}".format(i))
                sink.write({
                    'geometry': mapping(geom),
                    'properties': { 'id': int(i), 'station': row['m'], 'from_node': row['edge'][0], 'to_node': row['edge'][1]}
                })
    click.echo('Output written to: {}'.format(output_f))




import math
import bisect
from operator import itemgetter

from shapely.geometry import Point, LineString
from shapely.prepared import prep
import pandas as pnd

import surficial

def measure(line, start=0.0):
    """Return an array of vertex distances along a LineString.

    Parameters:
        line (LineString): the line on which to project.

    Other Parameters:
        start (float): measure at first vertex, zero by default.

    Returns:
        measures (list of float): list of vertex distances along line

    """
    measures = []
    prev = (None, None)
    for i, vert in enumerate(line.coords):
        if i == 0:
            d = start
        else:
            d += math.sqrt((vert[0]-prev[0])**2 + (vert[1]-prev[1])**2)
        measures.append(d)
        prev = vert
    return measures


def filter_contains(points, polygon):
    """Return a set of Points contained within a Polygon.

    Parameters:
        points (Point array): an array of Point to test.
        polygon (Polygon): the polygon to filter on.

    Returns:
        contained (list of Point): points contained within polygon

    """
    prepared_polygon = prep(polygon)
    contained = filter(prepared_polygon.contains, points)
    return contained


def project2d(point, line, measure=None):
    """Project a Point onto a line.

    Uses Shapely project(), which sets distance to zero for all negative distances.
    
    Note:
        The measure parameter does nothing at this time.
        I had intended to allow interpolation between measures other than distance.

    Parameters:
        point (Point): point at zero distance on line between point and p2.
        line (LineString): the line on which to project.

    Returns:
        result (dict): the projected Point, distance along line, offset from line.

    """
    m = line.project(point, normalized=False)
    projected_2d = line.interpolate(m)
    pt = Point([projected_2d.x, projected_2d.y, point.z])

    # locate the segment where the projected point lies
    if measure is None:
        measure = measure(line)
    i = bisect.bisect_left(measure, m)
    # should check first and last to avoid out of index
    coords = list(line.coords)
    d = orient2d(point, pt, Point(coords[i]), Point(coords[i-1]))

    result = {'pt':pt, 'm':m, 'd':d}
    return result


def orient2d(point, projection, from_vert, to_vert):
    """Calculate the orientation and offset distance of a point from a line.

    Parameters:
        point (Point): point for which we want to determine orientation left or right of a line
        projection (Point): point of projection onto the line
        from_vert (Point): from point defining the line
        to_vert (Point): to point defining the line

    Returns:
        offset (float): point distance offset from the line; negative is left of the line, positive or zero is right of the line

    """
    if (point.y - from_vert.y) * (to_vert.x - from_vert.x) - (point.x - from_vert.x) * (to_vert.y - from_vert.y) < 0:
        offset = -point.distance(projection) # the point is offset left of the line
    else:
        offset = point.distance(projection) # the point is offset right of the line
    return offset

def linestring_to_vertices(line):
    """Return a list of [m,x,y,z] values for a LineString

    :m: measure of distance along the line from the first vertex
    :x: vertex x coordinate
    :y: vertex y coordinate
    :z: vertex z coordinate

    Parameters:
        line (LineString): shapely LineString

    Returns:
        vertices (list): list of [m,x,y,z] values  

    """
    vertices = []
    for p in list(line.coords):
        position = line.project(Point(p))
        if len(p) == 3:
            vertices.append([position, p[0], p[1], p[2]])
        else:
            vertices.append([position, p[0], p[1], None])
    return vertices

def linestring_to_stations(line, position=0.0, step=1.0):
    """Return a list of regularly spaced stations along a LineString

    Parameters:
        line (LineString): shapely LineString 

    Other Parameters:
        position (float): distance along the line from the first vertex; permits stationing to begin at some offset from the first vertex
        step (float): distance in-between stations along the line

    Returns:
        stations (list): list of [m,x,y,z] values

    """
    stations = []
    while position < line.length:
        p = line.interpolate(position)
        if p.has_z:
            stations.append([position, p.x, p.y, p.z])
        else:
            stations.append([position, p.x, p.y, None])
        position += step
    return stations

def densify_linestring(line, start=0, step=10):
    """Densify a LineString with regularly-spaced stations

    Parameters:
        line (LineString): shapely LineString

    Other Parameters:
        start (float): distance along the line from the first vertex; permits stationing to begin at some offset from the first vertex
        step (float): distance in-between stations along the line

    Returns:
        dense_line (LineString): densified line

    """
    vertices = linestring_to_vertices(line)
    stations = linestring_to_stations(line, position=start, step=step)
    dense_vertices = sorted(vertices + stations, key=itemgetter(0), reverse=False)
    if line.has_z:
        dense_line = LineString([Point(x[1], x[2], x[3]) for x in dense_vertices])
    else:
        dense_line = LineString([Point(x[1], x[2]) for x in dense_vertices])

    return dense_line

"""
def densify_linestring_alt(line, start=0, step=10):
    position = 0
    for vertex in vertices[1:]:
        how_far = p1.distance(p2)
        while how_far > 0:
           if position == 0:
               add the first vertex
           if position < start & start < how_far:
               make a start vertex
               position += start
               how_far -= start
           if how_far > step:
               make a step vertex
               position += step
               how_far -= step
           else:
               add next vertex
               position += how_far
               how_far = 0
    return dense_line
"""



import sys
import logging
from pkg_resources import iter_entry_points

import click
from click_plugins import with_plugins

import surficial

logger = logging.getLogger(__name__)

@with_plugins(iter_entry_points('surficial.subcommands'))
@click.option('-v', '--verbose', default=False, is_flag=True, help="Enables verbose mode")
@click.version_option(version=surficial.__version__, message='%(version)s')
@click.group()
@click.pass_context
def cli(ctx, verbose):
    ctx.obj = {}
    ctx.obj['verbose'] = verbose
    if verbose:
        logging.basicConfig(stream=sys.stderr, level=logging.INFO)
    else:
        logging.basicConfig(stream=sys.stderr, level=logging.ERROR)


from setuptools import setup, find_packages


# Parse the version from the shapely module
for line in open('surficial/__init__.py', 'r'):
    if line.find("__version__") >= 0:
        version = line.split("=")[1].strip()
        version = version.strip('"')
        version = version.strip("'")
        continue

#open('VERSION.txt', 'wb').write(bytes(version, 'UTF-8'))
with open('VERSION.txt', 'w') as fp:
    fp.write(version)


setup(name='surficial',
	version=version,
	author='Michael Rahnis',
	author_email='michael.rahnis@fandm.edu',
	description='Python library and CLI tools to support analysis of stream long-profiles',
	url='http://github.com/mrahnis/surficial',
	license='BSD',
	packages=find_packages(),
	install_requires=[
		'click',
		'pandas',
		'matplotlib',
		'shapely',
		'descartes',
		'rasterio',
		'fiona',
		'networkx',
		'drapery'
	],
	entry_points='''
		[console_scripts]
		surficial=surficial.cli.surficial:cli

		[surficial.subcommands]
		buffer=surficial.cli.buffer:buffer
		network=surficial.cli.network:network
		plan=surficial.cli.plan:plan
		profile=surficial.cli.profile:profile
		repair=surficial.cli.repair:repair
		station=surficial.cli.station:station
	''',
	keywords='cross-section, topography, survey, plotting',
	classifiers=[
		'Development Status :: 3 - Alpha',
		'License :: OSI Approved :: BSD License',
		'Programming Language :: Python',
		'Programming Language :: Python :: 3',
		'Topic :: Scientific/Engineering :: GIS'
	]
)




import sys
import operator
from collections import Counter

import click
import fiona
from shapely.geometry import Point, LineString, shape, mapping

import surficial

def scan(test_point, points, decimal):
    for point in points:
        if Point(test_point[2]).almost_equals(Point(point[2]), decimal=decimal):
            yield point

def edit_line(line, edits):
    edit_line_ids = [edit[0] for edit in edits]
    geom = shape(line[1])
    if line[0] in edit_line_ids:
        for edit in edits:
            if line[0] == edit[0]:
                # could pop the edit at this point
                click.echo("Snapping id {0}, {1}, to: {2}".format(edit[0], edit[1], edit[3]))
                coords = list(geom.coords)
                if edit[1] == 'start':
                    geom = LineString([edit[3]] + coords[1:])
                elif edit[1] == 'end':
                    geom = LineString(coords[:-1] + [edit[3]])
                else:
                    print("Snap operation must be on start or end point")
    return geom


@click.command(options_metavar='<options>')
@click.argument('alignment_f', nargs=1, type=click.Path(exists=True), metavar='<alignment_file>')
@click.option('-o', '--output', 'output_f', nargs=1, type=click.Path(), metavar='<output_file>', help="Output file")
@click.option('-d', '--decimal', nargs=1, type=click.INT, metavar='<int>', default=6, help="Decimal place precision")
@click.pass_context
def repair(ctx, alignment_f, output_f, decimal):
    """
    Closes gaps in a network graph

    \b
    Example:
    surficial repair stream_ln.shp stream_ln_snap.shp --decimal 4

    """
    with fiona.open(alignment_f) as alignment_src:
        lines = [[line['id'], shape(line['geometry']), line['properties']] for line in alignment_src]
        source_driver = alignment_src.driver
        source_crs = alignment_src.crs
        source_schema = alignment_src.schema

    # make a list of [id, start/end, coords]
    endpoints = []
    for line in lines:
        endpoints.append([line[0], 'start', line[1].coords[0]])
        endpoints.append([line[0], 'end', line[1].coords[-1]])

    # find clusters endpoints to a decimal place precision
    clusters = []
    while endpoints:
        test_point = endpoints.pop(0)
        near_points = list(scan(test_point, endpoints, decimal))
        if len(near_points) > 0:
            # could broaden search by scanning each near_point but
            # just trying to fix failed snap points so pop the near_points
            cluster = [test_point] + near_points
            clusters.append(cluster)
            for i, point in enumerate(endpoints):
                if point in near_points:
                    endpoints.pop(i)

    # compile edits to snap endpoints to the most frequently occurring within a cluster
    edits = []
    for cluster in clusters:
        coords = [endpoint[2] for endpoint in cluster]
        keys = list(Counter(coords).keys())
        values = list(Counter(coords).values())
        index, value = max(enumerate(values), key=operator.itemgetter(1))
        snap_point = keys[index]

        for endpoint in cluster:
            if endpoint[2] != snap_point:
                edit = endpoint[:]
                edit.append(snap_point)
                edits.append(edit)

    # make the edits while writing out the data
    if output_f:
        with fiona.open(
            output_f,
            'w',
            driver=source_driver,
            crs=source_crs,
            schema=source_schema) as sink:
                for line in lines:
                    geom = edit_line(line, edits)
                    sink.write({
                        'geometry': mapping(geom),
                        'properties': line[2],
                    })
        click.echo('Completed, output written to: {}'.format(output_f))
    else:
        click.echo('No output file given, starting dry-run')
        for line in lines:
            edit_line(line, edits)
        click.echo('Completed')


import click
import fiona
from shapely.geometry import shape
import networkx as nx
import matplotlib.pyplot as plt

import surficial

@click.command(options_metavar='<options>')
@click.argument('alignment_f', nargs=1, type=click.Path(exists=True), metavar='<alignment_file>')
@click.pass_context
def network(ctx, alignment_f):
    """
    Plots the network graph

    \b
    Example:
    surficial network stream_ln.shp

    """
    with fiona.open(alignment_f) as alignment_src:
        lines = [shape(line['geometry']) for line in alignment_src]

    graph = surficial.Alignment(lines)

    # plot
    fig = plt.figure()
    ax = fig.add_subplot(111)
    nx.draw(graph, ax=ax, with_labels=True, node_color='w')

    plt.show()
